/* execute.c -- OR1K architecture dependent simulation
   Copyright (C) 1999 Damjan Lampret, lampret@opencores.org

This file is part of OpenRISC 1000 Architectural Simulator.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* Most of the OR1K simulation is done in here. */
/* 03/13/2000
   this version includes h.xxx/lxxx insns which could align at the address of multiple of 2
 */
/* OR16 ISA is added 03/13/2000, by Jimmy Chen-Min Chen, jimmy@ee.nctu.edu.tw 
				    Johan Rydberg, johan.rydberg@netinsight.se
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "arch.h"

#include "branch_predict.h"
#include "abstract.h"
#include "parse.h"
#include "trace.h"
#include "execute.h"
#include "stats.h"

/* General purpose registers. */
machword reg[MAX_GPRS];

/* Instruction queue */
struct iqueue_entry iqueue[20];

/* Benchmark multi issue execution */
int multissue[20];
int supercycles;

/* Completion queue */
struct icomplet_entry icomplet[20];

/* Program counter */
unsigned long pc;

/* Temporary program counter */
unsigned long pctemp;

/* OV: OVerflow flag */
int OV=0;

/* CARRY: carry flag */
int CARRY=0;

/* CCR */
int flag=0;

/* MAC: multiply-accumulate result registers */
int MACHI=0, MACLO=0;

/* CCR (for dependency calculation) */
char ccr_flag[10] = "flag";

/* Cycles counts fetch stages */
int cycles;

/* jump/branch flag to indicate the execution of 
   branch delay slot */
int jpbr_flag=0;
int no_delay_slot=0;
int branch_not_taken=0;
/* Implementation specific.
   Get an actual value of a specific register. */

machword eval_reg(char *regstr)
{
	int regno;
	
	debug("eval_reg(%s)\n", regstr);
	if (regstr && (*regstr == '%'))
		regno = atoi(regstr + 2);
	else
		regno = atoi(regstr + 1);

	if (regno < MAX_GPRS)
		return reg[regno];
	else {
		printf("\nEXCEPTION: read out of registers\n");
		cont_run = 0;
		return 0;
	}
}

/* Implementation specific.
   Set a specific register with value. */

void set_reg32(char *regstr, unsigned long value)
{
	int regno;

#if 0	
	if (strcmp(regstr, FRAME_REG) == 0) {
		printf("FP (%s) modified by insn at %x. ", FRAME_REG, pc);
		printf("Old:%.8lx  New:%.8lx\n", eval_reg(regstr), value);
	}

	if (strcmp(regstr, STACK_REG) == 0) {
		printf("SP (%s) modified by insn at %x. ", STACK_REG, pc);
		printf("Old:%.8lx  New:%.8lx\n", eval_reg(regstr), value);
	}
#endif
	if (regstr && (*regstr == '%'))
		regno = atoi(regstr + 2);
	else
		regno = atoi(regstr + 1);

	if (regno == 0)		/* gpr0 is always zero */
		value = 0;
	
	if (regno < MAX_GPRS)
		reg[regno] = value;
	else {
		printf("\nEXCEPTION: write out of registers\n");
		cont_run = 0;
	}
	return;
}

/* Does srcoperand depend on computation of dstoperand? Return
   non-zero if yes.

 Cycle t		 Cycle t+1
dst: irrelevant		src: immediate			always 0
dst: reg1 direct	src: reg2 direct		0 if reg1 != reg2
dst: reg1 disp		src: reg2 direct		always 0
dst: reg1 direct	src: reg2 disp			0 if reg1 != reg2
dst: reg1 disp		src: reg2 disp 			always 1 (store must
							finish before load)
*/

int depend_operands(char *dstoperand, char *srcoperand)
{
	char dst[OPERANDNAME_LEN];
	char src[OPERANDNAME_LEN];
	
	if (!srcoperand)
		return 0;

	if (!dstoperand)
		return 0;
		
	strcpy(dst, dstoperand);
	strcpy(src, srcoperand);

	if (*src == '#')		/* immediate */
		return 0;
	else
	if (strstr(src, "LO(") || strstr(src, "HI("))
		return 0;
	else
	if (!strstr(src, "(")) 
		if (*src == '%') {	/* src: reg direct */
			if (!strstr(dst, "("))
				if (*dst == '%')
					if (strcmp(dst, src) == 0)
						return 1; /* dst: reg direct */
					else
						return 0; /* dst: reg direct */
				else
					return 0; /* dst: addr */
			else
				return 0; /* dst: reg disp */
		} else
			return 0;	/* src: addr */
	else {				/* src: register disp */
		char *regstr;
		
		regstr = strstr(src, "(%r") + 1;	/* regstr == "rXX)" */
		*strstr(regstr, ")") = '\0';		/* regstr == "rXX" */

		if (!strstr(dst, "("))
			if (*dst == '%')
				if (strcmp(dst, regstr) == 0)
					return 1; /* dst: reg direct */
				else
					return 0; /* dst: reg direct */
			else
				return 0; /* dst: addr */
		else
			return 1; /* dst: reg disp */
	}	

	return 0;
}

/* Implementation specific.
   Get an actual value represented by operand (register direct, register
   indirect (with displacement), immediate etc.).
   
   #n		- immediate n
   rXX		- register direct
   XX		- relative or absolute address (labels)
   n(XX) 	- register indirect (with displacement) */
   
machword eval_operand(char *srcoperand)
{
	char operand[OPERANDNAME_LEN];
	
	strcpy(operand, srcoperand);

	if (*operand == '#')		/* immediate */
		return strtoul(&operand[1], NULL, 0);
	else
	if (((*operand == '-') || isdigit(*operand)) && !strstr(operand, "(")) {
		/* immediate */
		return strtoul(operand, NULL, 0);
	} else
	if (strncmp(operand, "LO(", 3) == 0) {
		*strchr(operand, ')') = '\0';
		if ((operand[3] == '-') || isdigit(operand[3]))
			return (unsigned long)strtol(&operand[3], NULL, 0) & 0xffff;
		else
			return eval_operand(&operand[3]) & 0xffff;
	} else 
	if (strncmp(operand, "HI(", 3) == 0) {
		*strchr(operand, ')') = '\0';
		if ((operand[3] == '-') || isdigit(operand[3]))
			return (unsigned long)strtol(&operand[3], NULL, 0) >> 16;
		else
			return eval_operand(&operand[3]) >> 16;
	} else 
	if (!strstr(operand, "(")) 	/* not indirect but ...*/
		if (*operand == '%') 	/* ... register direct */
			return eval_reg(operand);
		
		else 			/* ... rel. or abs. address */
			return eval_label(operand);
	else {				/* register indirect */
		int disp;		/* with possible displacement */
		char *regstr;
		unsigned int memaddr;
		
		/* disp = atoi(operand);	operand == "nn(rXX)" */
		disp = strtol(operand, NULL, 0);

		debug("eval_operand: disp=%u");
		regstr = strstr(operand, "(%r") + 1;	/* regstr == "rXX)" */
		*strstr(regstr, ")") = '\0';		/* regstr == "rXX" */
		debug("eval_operand: regstr=%s", regstr);
		memaddr = eval_reg(regstr) + disp;
		
		return eval_mem32(memaddr);
	}
	
	return 0;
}

/* Implementation specific.
   Set destination operand (register direct, register indirect
   (with displacement) with value. */
   
void set_operand(char *dstoperand, unsigned long value)
{
	char operand[OPERANDNAME_LEN];
	
	debug("set_operand %s <= %u\n", dstoperand, value);
	strcpy(operand, dstoperand);

	if (*operand == '#') {		/* immediate */
		printf("INTERNAL ERROR: Can't set immediate operand.\n");
		cont_run = 0;
	}
	else
	if (!strstr(operand, "(")) 	/* not indirect but ...*/
		if (*operand == '%') 	/* ... register direct */
			set_reg32(operand, value);
		else {			/* ... rel. or abs. address */
			/* printf("INTERNAL ERROR: Can't set addr operand.\n"); 
			cont_run = 0; */
			set_mem32(eval_label(operand), value);
		}
	else {				/* register indirect */
		int disp;		/* with possible displacement */
		char *regstr;
		unsigned int memaddr;
		char disp_str[10];
		
	#if 0
		regstr = strstr(operand, "(");
		strncpy(disp_str, operand, (regstr - operand));
		
	#endif
		/* disp = atoi(operand);	operand == "nn(rXX)" */
		disp = strtol(operand, NULL, 0);
	
		printf("\n XXXXXX disp = %x", disp);
		regstr = strstr(operand, "(%r") + 1;	/* regstr == "rXX)" */
		*strstr(regstr, ")") = '\0';		/* regstr == "rXX" */
		memaddr = eval_reg(regstr) + disp;
		
		set_mem32(memaddr, value);
	}

	return;
}

void reset()
{
	cycles = 0;
	supercycles = 0;
	memset(reg, 0, sizeof(reg));
	memset(iqueue, 0, sizeof(iqueue));
	memset(icomplet, 0, sizeof(icomplet));
	pctemp = eval_label("_main");
	pc = pctemp;
	debug("reset ...");
	set_reg32(STACK_REG, MEMORY_LEN - STACK_SIZE);
}

void fetch()
{

	char insn_first2_char[3];
	int h_insn_is_word_flag=0;
        char insn[INSNAME_LEN];

	if (!mem[pc].insn) {
		printf("exception: illegal instruction at 0x%x\n", pc);
		cont_run = 0;
		return;
	}
                                                                
#if 0
	pc = pctemp; /* pctemp : next program counter */
	pctemp += (h_insn_is_word_flag == 1) ? 2 : 4;
	printf("XXX nxt pc=%x\n", pc);
#endif

	/* Cycles after reset. */
	cycles++;

	/* Fetch instruction. */
	if(no_delay_slot) {
		printf("*** NOP is inherently loaded for no delay slot in h.bf/h.bnf insn ***\n");
		strcpy(iqueue[0].insn, "h.nop");
		strcpy(iqueue[0].op1, "");
		strcpy(iqueue[0].op2, "");
		strcpy(iqueue[0].op3, "");
		strcpy(iqueue[0].op4, "");
	}
	else {
		strcpy(iqueue[0].insn, mem[pc].insn->insn);
		strcpy(iqueue[0].op1, mem[pc].insn->op1);
		strcpy(iqueue[0].op2, mem[pc].insn->op2);
		strcpy(iqueue[0].op3, mem[pc].insn->op3);
		strcpy(iqueue[0].op4, mem[pc].insn->op4);
	}

	if(no_delay_slot) {
		/* re-show h.bf & h.bnf insn */
		iqueue[0].insn_addr = pc-2; /* because h.bf&h.bnf must be 16-bit insn */
	} else {
		iqueue[0].insn_addr = pc;
	}
	
	iqueue[0].dependdst = NULL;
	iqueue[0].dependsrc1 = NULL;
	iqueue[0].dependsrc2 = NULL;

	/* Increment program counter. */
#if 1
	strcpy(insn, mem[pc].insn->insn);

        insn_first2_char[0]=insn[0];
        insn_first2_char[1]=insn[1];
        insn_first2_char[2]='\0';

        if(strcmp("h.", insn_first2_char) == 0) {
                if(strcmp("h.load32u", insn) == 0 ||
                   strcmp("h.load16u", insn) == 0 ||
                   strcmp("h.load8u", insn) == 0 ||
                   strcmp("h.stor32", insn) == 0 ||
                   strcmp("h.stor16", insn) == 0 ||
                   strcmp("h.stor8", insn) == 0 ||
                   strcmp("h.jal", insn) == 0 || 
                   /* strcmp("h.mtsr", insn) == 0 || 
                   strcmp("h.mfsr", insn) == 0 || */
                   strcmp("h.movi16ze", insn) == 0 ||
                   strcmp("h.immhi16u", insn) == 0 ||
                   strcmp("h.addi16s", insn) == 0 ||
                   strcmp("h.subi16s", insn) == 0 ||
                   strcmp("h.xori16", insn) == 0 ||
                   strcmp("h.ori16", insn) == 0 ||
                   strcmp("h.andi16", insn) == 0
                  )
                        h_insn_is_word_flag = 2; /* h.xxx insn AND occupy 4 bytes */
                else
                        h_insn_is_word_flag = 1; /* h.xxx insn AND occupy 2 bytes */
        }
        else {
                        h_insn_is_word_flag = 0; /* not h.xxx insn */
        }
#endif

	/* Check for breakpoint. */
	if (mem[pc].brk)
		cont_run = 0;	/* Breakpoint set. */
		
	if(jpbr_flag != 0)
		jpbr_flag -= 1;
	if((cycles != 1) && (jpbr_flag == 0 || jpbr_flag == 2)) {
		if(h_insn_is_word_flag == 1)
			pctemp -= 2; /* if it's a 2-byte insn in delay slot */
	}
	printf("OOO pc=%x, pctemp=%x, branch_not_taken=%x\n", pc, pctemp, branch_not_taken);
	if(branch_not_taken == 0) { /* branch taken or not h.bf & h.bnf : don't update pc */
		pc = pctemp; /* pctemp : next program counter */
		/* pctemp += (h_insn_is_word_flag == 1) ? 2 : 4; */
		pctemp += 4;
	}
	printf("OOO jpbr_flag=%x, h_insn_is_word_flag=%x\n", jpbr_flag, h_insn_is_word_flag);
	printf("OOO current OP=%s, nxt PC=0x%x, nxt/nxt PC pctemp=0x%x", insn, pc, pctemp);

	return;
}

void decode(struct iqueue_entry *cur)
{

	cur->dependdst = cur->op1;
	cur->dependsrc1 = cur->op2;	/* for calculating register */
	cur->dependsrc2 = cur->op3;	/* dependency */
	
	cur->func_unit = unknown;
	no_delay_slot = 0;
	branch_not_taken = 0;
	
	if (strcmp(cur->insn, "h.nop") == 0) {
		cur->func_unit = nop;	
	} else

	if (strcmp(cur->insn, "h.sfeq32") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = eval_operand(cur->op1) == eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sfne32") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = eval_operand(cur->op1) != eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sfgt32s") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) >
		       (signed)eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sfge32s") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) >=
		       (signed)eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sflt32s") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) <
		       (signed)eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sfle32s") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) <=
		       (signed)eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sfgt32u") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) >
		       (unsigned)eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sfge32u") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) >=
		       (unsigned) eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sflt32u") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) <
		       (unsigned)eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.sfle32u") == 0) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) <=
		       (unsigned)eval_operand(cur->op2);
	} else

	if (strcmp(cur->insn, "h.ext16s") == 0) {
		cur->func_unit = extend;
		if ((eval_operand(cur->op1) & 0x8000) == 0x8000)
			set_operand(cur->op1, eval_operand(cur->op1) | 0xffff0000);
		else
			set_operand(cur->op1, eval_operand(cur->op1) & 0x0000ffff);
	} else
	if (strcmp(cur->insn, "h.ext8s") == 0) {
		cur->func_unit = extend;
		if ((eval_operand(cur->op1) & 0x80) == 0x80)
			set_operand(cur->op1, eval_operand(cur->op1) | 0xffffff00);
		else
			set_operand(cur->op1, eval_operand(cur->op1) & 0x000000ff);
	} else
	if (strcmp(cur->insn, "h.ext16z") == 0) {
		cur->func_unit = extend;
		set_operand(cur->op1, eval_operand(cur->op1) & 0x0000ffff);
	} else
	if (strcmp(cur->insn, "h.ext8z") == 0) {
		cur->func_unit = extend;
		set_operand(cur->op1, eval_operand(cur->op1) & 0x000000ff);
	} else

	if (strcmp(cur->insn, "h.load32u") == 0) {
		cur->func_unit = load;
		set_operand(cur->op1, eval_operand(cur->op2));
	} else
	if (strcmp(cur->insn, "h.load16u") == 0) {
		unsigned short temp = (eval_operand(cur->op2) >> 16);
		printf("OOO XXX load16u temp=%x", temp);
		cur->func_unit = load;
		set_operand(cur->op1, temp);
	} else
	if (strcmp(cur->insn, "h.load8u") == 0) {
		unsigned char temp = (eval_operand(cur->op2) >> 24);
		printf("OOO XXX load8u temp=%x", temp);
		cur->func_unit = load;
		set_operand(cur->op1, temp);
	} else

	if (strcmp(cur->insn, "h.stor32") == 0) {
		cur->func_unit = store;
		set_operand(cur->op1, eval_operand(cur->op2));
	} else
	if (strcmp(cur->insn, "h.stor16") == 0) {
		cur->func_unit = store;
		set_operand(cur->op1, (eval_operand(cur->op2) << 16) + (eval_operand(cur->op1) & 0xffff));
	} else
	if (strcmp(cur->insn, "h.stor8") == 0) {
		cur->func_unit = store;
		set_operand(cur->op1, (eval_operand(cur->op2) << 24) + (eval_operand(cur->op1) & 0xffffff));
	} else
	if (strcmp(cur->insn, "h.jmp") == 0) {
               	char *endptr;

		cur->func_unit = jump;
		jpbr_flag = 2; /* for half-word insn */
		/* printf("XXX jmp label is --%s--", cur->op1); */
		
               	strtol(cur->op1, &endptr, 0);
               	if (*endptr == '\0') {
               		debug("\nh.jmp relative: pc=%x pctemp=%x\n", pc, pctemp);
               		pctemp = pc + (signed)eval_operand(cur->op1) - 2;
               	} else
                        pctemp = eval_operand(cur->op1);
	} else
	if (strcmp(cur->insn, "h.jal") == 0) {
               	char *endptr;

		cur->func_unit = jump;
		jpbr_flag = 2; /* for half-word insn */

               	strtol(cur->op1, &endptr, 0);
               	if (*endptr == '\0') {
               		debug("\nh.jal relative: pc=%x pctemp=%x\n", pc, pctemp);
               		pctemp = pc + (signed)eval_operand(cur->op1) - 4;
               	} else
                        pctemp = eval_operand(cur->op1);

                set_reg32(LINK_REG, pc + 4);
	} else
	if (strcmp(cur->insn, "h.jalr") == 0) {
		cur->func_unit = jump;
		jpbr_flag = 2; /* for half-word insn */
		pctemp = eval_operand(cur->op1);
                set_reg32(LINK_REG, pc + 4);
	} else
	if (strcmp(cur->insn, "h.jr") == 0) {
		cur->func_unit = jump;
		jpbr_flag = 2; /* for half-word insn */
		cur->dependsrc1 = cur->op1;
		pctemp = eval_operand(cur->op1);
	} else

        if (strcmp(cur->insn, "h.bf") == 0) {
                cur->func_unit = branch;
                jpbr_flag = 2; /* for half-word insn */
                cur->dependsrc1 = ccr_flag;
		no_delay_slot = 1;
		if (eval_operand(cur->op1) >= pc)
			mstats.bnez.forward++;
		else
			mstats.bnez.backward++;
                if (flag) {
                	char *endptr;
                	
                	strtol(cur->op1, &endptr, 0);
                	if (*endptr == '\0') {
                		debug("\nh.bf relative: pc=%x pctemp=%x\n", pc, pctemp);
                		pctemp = pc + (signed)eval_operand(cur->op1) - 2;
                	} else
	                        pctemp = eval_operand(cur->op1);
                        mstats.bnez.taken++;
                        bpb_update(cur->insn_addr, 1);
                        btic_update(pctemp);
                } else {
			branch_not_taken = 1;
                        mstats.bnez.nottaken++;
                        bpb_update(cur->insn_addr, 0);
                        btic_update(pc);
                }
        } else
        if (strcmp(cur->insn, "h.bnf") == 0) {
                cur->func_unit = branch;
                jpbr_flag = 2; /* for half-word insn */
                cur->dependsrc1 = ccr_flag;
		no_delay_slot = 1;
		if (eval_operand(cur->op1) >= pc)
			mstats.beqz.forward++;
		else
			mstats.beqz.backward++;
                if (flag == 0) {
                	char *endptr;
                	
                	strtol(cur->op1, &endptr, 0);
                	if (*endptr == '\0') {
                		debug("\nh.bnf relative: pc=%x pctemp=%x\n", pc, pctemp);
                		pctemp = pc + (signed)eval_operand(cur->op1) - 2;
                	} else
	                        pctemp = eval_operand(cur->op1);
                        mstats.beqz.taken++;
                        bpb_update(cur->insn_addr, 1);
                        btic_update(pctemp);
                } else {
			branch_not_taken = 1;
                        mstats.beqz.nottaken++;
                        bpb_update(cur->insn_addr, 0);
                        btic_update(pc);
                }
        } else
	if (strcmp(cur->insn, "h.illegal") == 0) {
		cur->func_unit = jump;
		no_delay_slot = 1;
		pctemp = 0x0700;
	} else
	if (strcmp(cur->insn, "h.brk") == 0) {
		cur->func_unit = jump; /* action: PC <- 0xD00 */
		cur->dependsrc1 = cur->op1;
		no_delay_slot = 1;
		pctemp = 0x0d00;
	} else
	if (strcmp(cur->insn, "h.sys") == 0) {
		cur->func_unit = jump; /* action: PC <- 0xC00 */
		cur->dependsrc1 = cur->op1;
		no_delay_slot = 1;
		pctemp = 0x0c00;
	} else
	if (strcmp(cur->insn, "h.rfe") == 0) {
		cur->func_unit = exception; /* enter user mode */
	} else
        if (strcmp(cur->insn, "h.ma32s") == 0) {
                signed int temp3, temp2, temp1;
		signed long temp0;
                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
		MACLO += temp2 * temp3; /* do accumulation */
		MACHI = 0;
        } else
        if (strcmp(cur->insn, "h.ms32s") == 0) {
                signed int temp3, temp2, temp1;
		signed long temp0;
                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
                temp1 = temp2 * temp3;
		MACLO -= temp1;
		MACHI = 0;
	#if 0
                temp0 = MACHI;
                temp0 <<= 32;
                temp0 += MACLO; /* let temp0 = {MACHI, MACLO} */
                temp0 -= temp1; /* do subtraction */
                MACLO = (int) 0xffffffff & ((long) 0x00000000ffffffff & temp0);
                MACHI = temp0 >> 32;
	#endif
        } else

	if (strcmp(cur->insn, "h.mtsr") == 0) {
		int regno;
	#if 0 
		regno = strtol(cur->op1 + 1, NULL, 0); /* special register number */
	#endif
		regno = atoi(cur->op1 + 1); /* Special register number */
		if(regno == 3)
			MACHI = eval_operand(cur->op2);
		else if(regno == 4)
			MACLO = eval_operand(cur->op2);
	} else
	if (strcmp(cur->insn, "h.mfsr") == 0) {
		int regno;
		regno = atoi(cur->op2 + 1);
		if(regno == 3)
			set_operand(cur->op1, MACHI);
		else if(regno == 4)
			set_operand(cur->op1, MACLO);
	} else

	if (strcmp(cur->insn, "h.mul32s") == 0) {
		signed long temp5, temp4;
		signed long long temp3, temp2;
		signed long long temp1;
		cur->func_unit = arith;
		temp5 = (signed long) eval_operand(cur->op2);
                temp4 = (signed long) eval_operand(cur->op1);
		temp3 = (signed long long) temp5; /* sign extension */
		temp2 = (signed long long) temp4;
                temp1 = temp2 * temp3;
		MACLO = (unsigned long) (temp1 & (unsigned long long) 0x00000000ffffffff);
		MACHI = temp1 >> 32;
                set_operand(cur->op1, temp2);
	} else
        if (strcmp(cur->insn, "h.mul32u") == 0) {
                unsigned long long temp3, temp2;
		unsigned long long temp1;
                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
                temp1 = temp2 * temp3;
		MACLO = temp1 & (unsigned long long) 0x00000000ffffffff;
		MACHI = temp1 >> 32;
                set_operand(cur->op1, temp2);
        } else
        if (strcmp(cur->insn, "h.div32s") == 0) {
                signed long temp3, temp2, temp1;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
		if(temp3 != 0) {
			temp1 = temp2 / temp3;
			set_operand(cur->op1, temp1);
		}
		else {
			printf("\nDIVISOR ZERO exception\n");
			exit(1);
		}
        } else
        if (strcmp(cur->insn, "h.div32u") == 0) {
                unsigned long temp3, temp2, temp1;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
		if(temp3 != 0) {
			temp1 = temp2 / temp3;
			set_operand(cur->op1, temp1);
		}
		else {
			printf("\nDIVISOR ZERO exception\n");
			exit(1);
		}
        } else
        if (strcmp(cur->insn, "h.mod32s") == 0) {
                signed long temp3, temp2, temp1;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
		temp2 = eval_operand(cur->op1);
		if(temp3 != 0) {	
			temp1 = temp2 % temp3;
			set_operand(cur->op1, temp1);
		}
		else {
			printf("\nDIVISOR ZERO exception\n");
			exit(1);
		}
        } else
        if (strcmp(cur->insn, "h.mod32u") == 0) {
                unsigned long temp3, temp2, temp1;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
		if(temp3 != 0) {
			temp1 = temp2 % temp3;
			set_operand(cur->op1, temp1);
		}
		else {
			printf("\nDIVISOR ZERO exception\n");
			exit(1);
		}
        } else

 	if (strcmp(cur->insn, "h.movi8se") == 0) {
		cur->func_unit = movimm;
		set_operand(cur->op1, (signed)eval_operand(cur->op2));
	} else
 	if (strcmp(cur->insn, "h.movi16ze") == 0) {
		cur->func_unit = movimm;
		set_operand(cur->op1, (0x0000ffff & eval_operand(cur->op2)) );
	} else
        if (strcmp(cur->insn, "h.mov32") == 0) {
                cur->func_unit = move;
                set_operand(cur->op1, eval_operand(cur->op2));
        } else
	if (strcmp(cur->insn, "h.immhi16u") == 0) {
		cur->func_unit = movimm;
		set_operand(cur->op1, (eval_operand(cur->op1) & 0xffff) | (eval_operand(cur->op2) << 16));
	} else

        if (strcmp(cur->insn, "h.add32s") == 0) {
                signed long long temp3, temp2;
		signed long long temp1;
                signed char temp4;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op3);
                temp2 = eval_operand(cur->op2);
                temp1 = temp2 + temp3;
		CARRY = 0x00000001 & (temp1 >> 32);

		OV = ~(0x0001 & temp1>>31) &  (0x0001 & (temp3>>31)) &  (0x0001 & (temp2>>31)) |
		      (0x0001 & temp1>>31) & ~(0x0001 & (temp3>>31)) & ~(0x0001 & (temp2>>31));

		if(OV) {
			no_delay_slot = 1;
			pctemp = 0x0E00;
		}
                else {
                	set_operand(cur->op1, temp1);
			temp4 = temp1;
			if (temp4 == temp1)
				mstats.byteadd++;
		}
        } else
        if (strcmp(cur->insn, "h.addwc32s") == 0) {
                signed long long temp3, temp2;
                signed long long temp1;
                signed char temp4;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
                temp1 = temp2 + temp3 + CARRY;
                CARRY = 0x00000001 & (temp1 >> 32);

		OV = ~(0x0001 & temp1>>31) &  (0x0001 & (temp3>>31)) &  (0x0001 & (temp2>>31)) |
		      (0x0001 & temp1>>31) & ~(0x0001 & (temp3>>31)) & ~(0x0001 & (temp2>>31));

                if(OV) {
                        no_delay_slot = 1;
                        pctemp = 0x0E00;
                }
                else {
                	set_operand(cur->op1, temp1);
                        temp4 = temp1;
                        if (temp4 == temp1)
                                mstats.byteadd++;
                }
        } else
        if (strcmp(cur->insn, "h.add32u") == 0) {
                unsigned long long temp3, temp2;
		unsigned long long temp1;
                unsigned char temp4;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
                temp1 = temp2 + temp3;
		CARRY = 0x00000001 & (temp1 >> 32);
                set_operand(cur->op1, temp1);

                temp4 = temp1;
                if (temp4 == temp1)
                        mstats.byteadd++;
        } else
        if (strcmp(cur->insn, "h.addi16s") == 0) {
                signed long long temp3, temp2;
		signed long long temp1;
                signed char temp4;

                cur->func_unit = arith;
                temp3 = eval_operand(cur->op3);
                temp2 = eval_operand(cur->op2);
                temp1 = temp2 + temp3;
		CARRY = 0x00000001 & (temp1 >> 32);

		OV = ~(0x0001 & temp1>>31) &  (0x0001 & (temp3>>31)) &  (0x0001 & (temp2>>31)) |
		      (0x0001 & temp1>>31) & ~(0x0001 & (temp3>>31)) & ~(0x0001 & (temp2>>31));
                if(OV) {
                        no_delay_slot = 1;
                        pctemp = 0x0E00;
                }
                else {
                	set_operand(cur->op1, temp1);
                        temp4 = temp1;
                        if (temp4 == temp1)
                                mstats.byteadd++;
                }
        } else

        if (strcmp(cur->insn, "h.subi16s") == 0) {
                signed long long temp3, temp2;
		signed long long temp1;
                signed char temp4;
                cur->func_unit = arith;
                temp3 = eval_operand(cur->op3);
                temp2 = eval_operand(cur->op2);
                temp1 = temp2 - temp3;
		CARRY = 0x00000001 & (temp1 >> 32);

		OV =  (0x0001 & temp1>>31) & ~(0x0001 & (temp3>>31)) &  (0x0001 & (temp2>>31)) |
		     ~(0x0001 & temp1>>31) &  (0x0001 & (temp3>>31)) & ~(0x0001 & (temp2>>31));

                if(OV) {
                        no_delay_slot = 1;
                        pctemp = 0x0E00;
                }
                else {
                	set_operand(cur->op1, temp1);
                        temp4 = temp1;
                        if (temp4 == temp1)
                                mstats.byteadd++;
                }
        } else
        if (strcmp(cur->insn, "h.sub32s") == 0) {
                signed long long temp3, temp2;
		signed long long temp1;
                signed char temp4;
                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
                temp1 = temp2 - temp3;
		CARRY = 0x00000001 & (temp1 >> 32);

		OV =  (0x0001 & temp1>>31) & ~(0x0001 & (temp3>>31)) &  (0x0001 & (temp2>>31)) |
		     ~(0x0001 & temp1>>31) &  (0x0001 & (temp3>>31)) & ~(0x0001 & (temp2>>31));
                if(OV) {
                        no_delay_slot = 1;
                        pctemp = 0x0E00;
                }
                else {
                	set_operand(cur->op1, temp1);
                        temp4 = temp1;
                        if (temp4 == temp1)
                                mstats.byteadd++;
                }
        } else
        if (strcmp(cur->insn, "h.subwc32s") == 0) {
                signed long long temp3, temp2;
                signed long long temp1;
                signed char temp4;
                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
                temp1 = temp2 - temp3 - CARRY;
                CARRY = 0x00000001 & (temp1 >> 32);

		OV =  (0x0001 & temp1>>31) & ~(0x0001 & (temp3>>31)) &  (0x0001 & (temp2>>31)) |
		     ~(0x0001 & temp1>>31) &  (0x0001 & (temp3>>31)) & ~(0x0001 & (temp2>>31));

                if(OV) {
                        no_delay_slot = 1;
                        pctemp = 0x0E00;
                }
                else {
                	set_operand(cur->op1, temp1);
                        temp4 = temp1;
                        if (temp4 == temp1)
                                mstats.byteadd++;
                }
        } else
        if (strcmp(cur->insn, "h.sub32u") == 0) {
                unsigned long long temp3, temp2;
		unsigned long long temp1;
                cur->func_unit = arith;
                temp3 = eval_operand(cur->op2);
                temp2 = eval_operand(cur->op1);
                temp1 = temp2 - temp3;
		CARRY = 0x00000001 & (temp1 >> 32);
                set_operand(cur->op1, temp1);
        } else

        if (strcmp(cur->insn, "h.xori16") == 0) {
                cur->func_unit = arith;
                set_operand(cur->op1, eval_operand(cur->op2) ^ (eval_operand(cur->op3) & 0xffff));
        } else
        if (strcmp(cur->insn, "h.xor32") == 0) {
                cur->func_unit = arith;
                set_operand(cur->op1, eval_operand(cur->op1) ^ eval_operand(cur->op2));
        } else
        if (strcmp(cur->insn, "h.ori16") == 0) {
                cur->func_unit = arith;
                set_operand(cur->op1, eval_operand(cur->op2) | (eval_operand(cur->op3) & 0xffff));
        } else
        if (strcmp(cur->insn, "h.or32") == 0) {
                cur->func_unit = arith;
                set_operand(cur->op1, eval_operand(cur->op1) | eval_operand(cur->op2));
        } else
        if (strcmp(cur->insn, "h.andi16") == 0) {
                cur->func_unit = arith;
                set_operand(cur->op1, eval_operand(cur->op2) & (eval_operand(cur->op3) & 0xffff));
        } else
        if (strcmp(cur->insn, "h.and32") == 0) {
                cur->func_unit = arith;
                set_operand(cur->op1, eval_operand(cur->op1) & eval_operand(cur->op2));
        } else

        if (strcmp(cur->insn, "h.shlai32") == 0) {
                signed long temp1;
                cur->func_unit = shift;
                /* debug("shla: op3:%d op4:%d", eval_operand(cur->op3), eval_operand(cur->o
                   debug("shla: op4=%s", cur->op4); */

                temp1 = eval_operand(cur->op1);
                temp1 <<= (0x001f & eval_operand(cur->op2));
                set_operand(cur->op1, temp1);
        } else
        if (strcmp(cur->insn, "h.shrai32") == 0) {
                signed long temp1;

                cur->func_unit = shift;

                /* operand 1 is to be shifted; operand 2 is shift count */

                temp1 = (signed) eval_operand(cur->op1);
                temp1 >>= (0x001f & eval_operand(cur->op2));
                set_operand(cur->op1, temp1);
        } else
        if (strcmp(cur->insn, "h.shrli32") == 0) {
                cur->func_unit = shift;
		set_operand(cur->op1, eval_operand(cur->op1) >> (0x001f & eval_operand(cur->op2)) );
        } else

        if (strcmp(cur->insn, "h.ror32") == 0) {
		unsigned int temp1, temp2, temp3;
		cur->func_unit = shift;
		
		temp1 = eval_operand(cur->op1);
		temp2 = temp1 >> (0x001f & eval_operand(cur->op2));
		temp3 = temp1 << (32 - (0x001f & eval_operand(cur->op2)));
		set_operand(cur->op1, temp2 | temp3);
	} else
        if (strcmp(cur->insn, "h.rol32") == 0) {
		unsigned int temp1, temp2, temp3;
		cur->func_unit = shift;
		
		temp1 = eval_operand(cur->op1);
		temp2 = temp1 << (0x001f & eval_operand(cur->op2));
		temp3 = temp1 >> (32 - (0x001f & eval_operand(cur->op2)));
		set_operand(cur->op1, temp2 | temp3);
	} else

        if (strcmp(cur->insn, "h.shla32") == 0) {
                signed long temp1;
                cur->func_unit = shift;
                /* debug("shla: op3:%d op4:%d", eval_operand(cur->op3), eval_operand(cur->op4));
                   debug("shla: op4=%s", cur->op4); */

                temp1 = eval_operand(cur->op1);
                temp1 <<= (0x001f & eval_operand(cur->op2));
                set_operand(cur->op1, temp1);
        } else
        if (strcmp(cur->insn, "h.shra32") == 0) {
                signed long temp1;

                cur->func_unit = shift;

                /* operand 1 is to be shifted; operand 2 is shift count */

                temp1 = (signed) eval_operand(cur->op1);
                temp1 >>= (0x001f & eval_operand(cur->op2));
                set_operand(cur->op1, temp1);
        } else
	if (strcmp(cur->insn, "h.shrl32") == 0) {
		cur->func_unit = shift;
		set_operand(cur->op1, eval_operand(cur->op1) >> (0x001f & eval_operand(cur->op2)) );
	} else

/* strange insn */
	if (strcmp(cur->insn, "simrdtsc") == 0) {
		set_operand(cur->op1, cycles);
	} else
	if (strcmp(cur->insn, "simprintf") == 0) {
		unsigned long stackaddr, fmtaddr, args;
		
		stackaddr = eval_reg(FRAME_REG);
		simprintf(stackaddr, eval_reg("r3"));
		debug("simprintf %x %x %x\n", stackaddr, fmtaddr, args);
/* Illegal opcode */
	} else {
		printf("\nEXCEPTION: illegal opcode %s ", cur->insn);
		printf("at %.8lx\n", cur->insn_addr);
	/* jump to exception service routine at 0x0700 */
		cur->func_unit = jump;
/*		pctemp = 0x0700;

		cont_run = 0; */
	}

	/* Dynamic, dependency stats. */
	adddstats(icomplet[0].insn, iqueue[0].insn, 1, check_depend());

	/* Dynamic, functional units stats. */
	addfstats(icomplet[0].func_unit, iqueue[0].func_unit, 1, check_depend());

	/* Dynamic, single stats. */
	addsstats(iqueue[0].insn, 1, 0);

	/* Pseudo multiple issue benchmark */
	if ((multissue[cur->func_unit] == 0) || (check_depend())) {
		int i;
		for (i = 0; i < 20; i++)
			multissue[i] = 1;
		supercycles++;
		/* if (check_depend())
			supercycles++; */
		multissue[move] = 2;
		multissue[movimm] = 2;
		multissue[arith] = 3;
		multissue[store] = 4;
		multissue[load] = 2;
	}
	multissue[cur->func_unit]--;
	
	return;
}

void execute()
{
	int i;
	
	/* Here comes real execution someday... */
	
	/* Instruction waits in completition buffer until retired. */
	strcpy(icomplet[0].insn, iqueue[0].insn);
	strcpy(icomplet[0].op1, iqueue[0].op1);
	strcpy(icomplet[0].op2, iqueue[0].op2);
	strcpy(icomplet[0].op3, iqueue[0].op3);
	strcpy(icomplet[0].op4, iqueue[0].op4);
	icomplet[0].func_unit = iqueue[0].func_unit;	
	icomplet[0].insn_addr = iqueue[0].insn_addr;

	if (iqueue[0].dependdst == iqueue[0].op1)
		icomplet[0].dependdst = icomplet[0].op1;
	else
	if (iqueue[0].dependdst == iqueue[0].op2)
		icomplet[0].dependdst = icomplet[0].op2;
	else
	if (iqueue[0].dependdst == iqueue[0].op3)
		icomplet[0].dependdst = icomplet[0].op3;
	else
		icomplet[0].dependdst = NULL;

	if (iqueue[0].dependsrc1 == iqueue[0].op1)
		icomplet[0].dependsrc1 = icomplet[0].op1;
	else
	if (iqueue[0].dependsrc1 == iqueue[0].op2)
		icomplet[0].dependsrc1 = icomplet[0].op2;
	else
	if (iqueue[0].dependsrc1 == iqueue[0].op3)
		icomplet[0].dependsrc1 = icomplet[0].op3;
	else
		icomplet[0].dependsrc1 = NULL;

	if (iqueue[0].dependsrc2 == iqueue[0].op1)
		icomplet[0].dependsrc2 = icomplet[0].op1;
	else
	if (iqueue[0].dependsrc2 == iqueue[0].op2)
		icomplet[0].dependsrc2 = icomplet[0].op2;
	else
	if (iqueue[0].dependsrc2 == iqueue[0].op3)
		icomplet[0].dependsrc2 = icomplet[0].op3;
	else
		icomplet[0].dependsrc2 = NULL;
	
	/* History of execution */
	for (i = HISTEXEC_LEN - 1; i; i--)
		histexec[i] = histexec[i - 1];
	histexec[0] = icomplet[0].insn_addr;	/* add last insn */
	
	return;
}

void dumpreg()
{
	int i;
	
	printf("\n\nIQ[0]:");
	dumpmemory(iqueue[0].insn_addr, iqueue[0].insn_addr + 4);
	printf(" (just executed)\tCYCLE: %u \tSUPERCYCLE: %u\nPC:", cycles, supercycles);
	if(no_delay_slot) {
	}
	else {
		dumpmemory(pc, pc + 4);
	}
	printf(" (next insn)");
	for(i = 0; i < MAX_GPRS; i++) {
		if (i % 4 == 0)
			printf("\n");
		printf("GPR%.2u: %.8lx  ", i, reg[i]);
	}
	/* printf("\nflag = %u, MACHI = %x, MACLO = %x\n", flag, MACHI, MACLO); */
	printf("\nflag = %u, CARRY = %u, OV = %u,  MACHI = %08x, MACLO = %08x\n", flag, CARRY, OV, MACHI, MACLO);
}
