/* decode.c -- Decode OR1K machine code to assembly. Handles OR32 and OR16.
   Copyright (C) 1999 Damjan Lampret, lampret@opencores.org

This file is part of OpenRISC 1000 Architectural Simulator.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* Following routine is the same as in objdump
   (binutils/opcodes/or[32,16]-dis.c) and it translates one machine
   code instruction to unassembled instruction. */

#if OR32 || OR16

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "arch.h"

#if OR32
#include "or32.h"
#endif
#if OR16
#include "or16.h"
#endif

#if OR32
#define machine_opcode or32_opcode
#define machine_opcodes or32_opcodes
#define machine_letter or32_letter
#define machine_letters or32_letters
#else
#define machine_opcode or16_opcode
#define machine_opcodes or16_opcodes
#define machine_letter or16_letter
#define machine_letters or16_letters
#endif

extern int cont_run;

char disassembled_str[100];

char *disassembled = &disassembled_str[0];

/* Some of the following routines are also in orX-dic.c and in tc-orX.c. We
   need to have a library or something for easier maintenance. */
   
/* Is individual insn's operand signed or unsigned? */
int letter_signed_OBSOLETE(char l)
{
        struct machine_letter *pletter;

        for(pletter = machine_letters; pletter->letter != '\0'; pletter++)
                if (pletter->letter == l)
                        return pletter->sign;

        printf("\nABORT: letter_signed(%c): Unknown letter.\n", l);
        exit(1);
}

/* Calculates instruction length in bytes. Either 2 or 4 for OR16
   and always 4 for OR32. */
int insn_len_OBSOLETE(char *insn)
{
        struct machine_opcode *pinsn;
        char *enc;
        int len = 0;
	
        for(pinsn = machine_opcodes; strlen(pinsn->name); pinsn++) {
                if (strcmp(pinsn->name, insn) == 0) {
                        for (enc = pinsn->encoding; *enc != '\0'; enc++)
                                if ((*enc == '0') && (*(enc+1) == 'x')) {
                                        len += 4;
                                        enc += 2;
                                }
                                else if (!isspace(*enc))
                                        len++;
                        return len / 8;
                }
        }
        printf("\nABORT: insn_len(%s): Unknown instruction.\n", insn);
        cont_run = 0;
        return 4;
}

/* Number of letters in the individual lettered operand. */
int letter_range_OBSOLETE(char l)
{
        struct machine_opcode *pinsn;
        char *enc;
        int range = 0;

        for(pinsn = machine_opcodes; strlen(pinsn->name); pinsn++) {
                if (strchr(pinsn->encoding,l)) {
                        for (enc = pinsn->encoding; *enc != '\0'; enc++)
                                if ((*enc == '0') && (*(enc+1) == 'x')) {
                                        enc += 2;
                                }
                                else if (*enc == l)
                                        range++;
                        return range;
                }
        }
        printf("\nABORT: letter_range(%c): Never used letter.\n", l);
        exit(1);
}

/* Automagically does zero- or sign- extension and also finds correct
   sign bit position if sign extension is correct extension. Which extension
   is proper is figured out from letter description. */
unsigned long extend_imm(unsigned long imm, char l)
{
	unsigned long mask;
	int letter_bits;
	
	/* First truncate all bits above valid range for this letter
	   in case it is zero extend. */
	letter_bits = letter_range(l);
	mask = (1 << letter_bits) - 1;
	imm &= mask;
	
	/* Do sign extend if this is the right one. */
	if (letter_signed(l) && (imm >> (letter_bits - 1)))
		imm |= (~mask);
	
	return imm;
}

static unsigned long
or16_extract(param_ch, enc_initial, insn)
     char param_ch;
     char *enc_initial;
     unsigned long insn;
{
  char *enc;
  unsigned long ret = 0;
  int opc_pos = 0;
  int param_pos = 0;

  for (enc = enc_initial; *enc != '\0'; enc++)
    if (*enc == param_ch)
      if (enc - 2 >= enc_initial && (*(enc - 2) == '0') && (*(enc - 1) == 'x'))
      	continue;
      else
        param_pos++;

#if DEBUG
  printf("or16_extract: %x ", param_pos);
#endif
  opc_pos = 32;
  for (enc = enc_initial; *enc != '\0'; )
    if ((*enc == '0') && (*(enc+1) == 'x')) 
      {
        opc_pos -= 4;
        if ((param_ch == '0') || (param_ch == '1')) 
          {
            unsigned long tmp = strtol(enc, NULL, 16);
#if DEBUG
            printf(" enc=%s, tmp=%x ", enc, tmp);
#endif
            if (param_ch == '0')
              tmp = 15 - tmp;
            ret |= tmp << opc_pos;
          }
        enc += 3;
      }
    else if ((*enc == '0') || (*enc == '1')) 
      {
        opc_pos--;
        if (param_ch == *enc)
          ret |= 1 << opc_pos;
        enc++;
      }
    else if (*enc == param_ch) 
      {
        opc_pos--;
        param_pos--;
#if DEBUG
        printf("\n  ret=%x opc_pos=%x, param_pos=%x\n", ret, opc_pos, param_pos);
#endif  
        if (islower(param_ch))
          ret -= ((insn >> opc_pos) & 0x1) << param_pos;
        else
          ret += ((insn >> opc_pos) & 0x1) << param_pos;
        enc++;
      }
    else if (isalpha(*enc)) 
      {
        opc_pos--;
        enc++;
      }
    else if (*enc == '-') 
      {
        opc_pos--;
        enc++;
      }
    else
      enc++;

#if DEBUG
  printf ("ret=%x\n", ret);
#endif
  return ret;
}

static int
or16_opcode_match (insn, encoding)
     unsigned long insn;
     char *encoding;
{
  unsigned long ones, zeros;

#if DEBUG
  printf ("or16_opcode_match: %.8lx\n", insn);
#endif    
  ones  = or16_extract('1', encoding, insn);
  zeros = or16_extract('0', encoding, insn);
  
#if DEBUG
  printf("ones: %x \n", ones);
  printf("zeros: %x \n", zeros);
#endif
  if ((insn & ones) != ones) 
    {
#if DEBUG
      printf("ret1\n");
#endif
      return 0;
    }
    
  if ((~insn & zeros) != zeros) 
    {
#if DEBUG
      printf("ret2\n");
#endif
      return 0;
    }
  
#if DEBUG
  printf ("ret3\n");
#endif
  return 1;
}

/* Print register. Used only by print_insn. */

static void
or16_print_register (param_ch, encoding, insn)
     char param_ch;
     char *encoding;
     unsigned long insn;
{
  int regnum = or16_extract(param_ch, encoding, insn);
  
#if DEBUG
  printf("or16_print_register: %c, %s, %x\n", param_ch, encoding, insn);
#endif  
#if OR16
   sprintf (disassembled, "%s%%r%d", disassembled, regnum);
#else
   sprintf (disassembled, "%sr%d", disassembled, regnum);
#endif
}

/* Print immediate. Used only by print_insn. */

static void
or16_print_immediate (param_ch, encoding, insn)
     char param_ch;
     char *encoding;
     unsigned long insn;
{
  int imm = or16_extract (param_ch, encoding, insn);

  imm = extend_imm(imm, param_ch);
  
  if (letter_signed(param_ch))
    sprintf (disassembled, "%s%d", disassembled, imm);
  else
    sprintf (disassembled, "%s%#x", disassembled, imm);
}

/* Disassemble one instruction from insn to disassemble.
   Return the size of the instruction (2 or 4 on or16).  */
int
disassemble_insn (insn)
     unsigned long insn;
{
  /* Address. Will be sign extened 12-bit */
  int addr;

  struct machine_opcode CONST * opcode;

  for (opcode = &machine_opcodes[0];
       opcode < &machine_opcodes[num_opcodes];
       ++opcode)
    {
      if (or16_opcode_match (insn, opcode->encoding))
        {
          char *s;
    
          sprintf(disassembled, "%s ", opcode->name);
          for (s = opcode->args; *s != '\0'; ++s)
            {
              switch (*s)
                {
                case '\0':
                  return 4;
      
                case 'r':
                  or16_print_register(*++s, opcode->encoding, insn);
                  break;

                default:
                  if (strchr (opcode->encoding, *s))
                    or16_print_immediate (*s, opcode->encoding, insn);
                  else
                    sprintf(disassembled, "%s%c", disassembled, *s);
                }
            }

	  return insn_len(opcode->name);
        }
    }

  /* This used to be %8x for binutils.  */
  sprintf(disassembled, "%s.word 0x%08x", disassembled, insn);
  return 4;
}

#else

char *disassembled = "";

/* No decoding for DLX and others. */
int
disassemble_insn (insn)
{
	return 4;
}

int insn_len(char *insn)
{
	return 4;
}

#endif
