/* execute.c -- OR1K architecture dependent simulation
   Copyright (C) 1999 Damjan Lampret, lampret@opencores.org

This file is part of OpenRISC 1000 Architectural Simulator.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* Most of the OR1K simulation is done in here. */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "arch.h"

#include "branch_predict.h"
#include "abstract.h"
#include "parse.h"
#include "trace.h"
#include "execute.h"
#include "stats.h"
#include "except.h"
#include "sprs.h"

/* General purpose registers. */
machword reg[MAX_GPRS];

/* Instruction queue */
struct iqueue_entry iqueue[20];

/* Benchmark multi issue execution */
int multissue[20];
int supercycles;
int issued_per_cycle = 4;
int hazardwait = 0;
int nops = 0;
int nop_period = 0;
int nop_maxperiod = 0;

/* Completition queue */
struct icomplet_entry icomplet[20];

/* Program counter */
unsigned long pc;

/* Temporary program counter */
unsigned long pctemp;

/* CCR */
int flag;

/* CCR (for dependency calculation) */
char ccr_flag[10] = "flag";

/* Cycles counts fetch stages */
int cycles;

/* Load and store stalls */
int loadcycles, storecycles;

/* Implementation specific.
   Get an actual value of a specific register. */

machword eval_reg(char *regstr)
{
	int regno;
	
	debug("atoi\n");
	regno = atoi(regstr + 1);

	if (regno < MAX_GPRS) {
		int delta = (cycles - raw_stats.reg[regno]);
		
		if (delta < RAW_RANGE)
			raw_stats.range[delta]++;
	debug("atoi ret1\n");
		return reg[regno];
	} else {
		printf("\nEXCEPTION: read out of registers\n");
		cont_run = 0;
	debug("atoi ret2\n");
		return 0;
	}
}

/* Implementation specific.
   Set a specific register with value. */

void set_reg32(char *regstr, unsigned long value)
{
	int regno;

#if 0	
	if (strcmp(regstr, FRAME_REG) == 0) {
		printf("FP (%s) modified by insn at %x. ", FRAME_REG, pc);
		printf("Old:%.8lx  New:%.8lx\n", eval_reg(regstr), value);
	}

	if (strcmp(regstr, STACK_REG) == 0) {
		printf("SP (%s) modified by insn at %x. ", STACK_REG, pc);
		printf("Old:%.8lx  New:%.8lx\n", eval_reg(regstr), value);
	}
#endif
	regno = atoi(regstr + 1);

	if (regno == 0)		/* gpr0 is always zero */
		value = 0;
	
	if (regno < MAX_GPRS) {
		reg[regno] = value;
		raw_stats.reg[regno] = cycles;
	} else {
		printf("\nEXCEPTION: write out of registers\n");
		cont_run = 0;
	}
	
	
	return;
}

/* Does srcoperand depend on computation of dstoperand? Return
   non-zero if yes.

 Cycle t		 Cycle t+1
dst: irrelevant		src: immediate			always 0
dst: reg1 direct	src: reg2 direct		0 if reg1 != reg2
dst: reg1 disp		src: reg2 direct		always 0
dst: reg1 direct	src: reg2 disp			0 if reg1 != reg2
dst: reg1 disp		src: reg2 disp 			always 1 (store must
							finish before load)
*/

int depend_operands(char *dstoperand, char *srcoperand)
{
	char dst[OPERANDNAME_LEN];
	char src[OPERANDNAME_LEN];
	
	if (!srcoperand)
		return 0;

	if (!dstoperand)
		return 0;
		
	strcpy(dst, dstoperand);
	strcpy(src, srcoperand);

	if (*src == '#')		/* immediate */
		return 0;
	else
	if (strstr(src, "lo(") || strstr(src, "hi("))
		return 0;
	else
	if (!strstr(src, "(")) 
		if (*src == 'r') {	/* src: reg direct */
			if (!strstr(dst, "("))
				if (*dst == 'r')
					if (strcmp(dst, src) == 0)
						return 1; /* dst: reg direct */
					else
						return 0; /* dst: reg direct */
				else
					return 0; /* dst: addr */
			else
				return 0; /* dst: reg disp */
		} else
			return 0;	/* src: addr */
	else {				/* src: register disp */
		char *regstr;
		
		regstr = strstr(src, "(r") + 1;	/* regstr == "rXX)" */
		*strstr(regstr, ")") = '\0';		/* regstr == "rXX" */

		if (!strstr(dst, "("))
			if (*dst == 'r')
				if (strcmp(dst, regstr) == 0)
					return 1; /* dst: reg direct */
				else
					return 0; /* dst: reg direct */
			else
				return 0; /* dst: addr */
		else
			return 1; /* dst: reg disp */
	}	

	return 0;
}

/* Implementation specific.
   Get an actual value represented by operand (register direct, register
   indirect (with displacement), immediate etc.).
   
   #n		- immediate n
   rXX		- register direct
   XX		- relative or absolute address (labels)
   n(XX) 	- register indirect (with displacement) */
   
machword eval_operand(char *srcoperand)
{
	char operand[OPERANDNAME_LEN];
	
	strcpy(operand, srcoperand);

	if (*operand == '#')		/* immediate */
		return strtoul(&operand[1], NULL, 0);
	else
	if (((*operand == '-') || isdigit(*operand)) && !strstr(operand, "(")) {
		/* immediate */
		return strtoul(operand, NULL, 0);
	} else
	if (strncmp(operand, "lo(", 3) == 0) {
		*strchr(operand, ')') = '\0';
		if ((operand[3] == '-') || isdigit(operand[3]))
			return (unsigned long)strtol(&operand[3], NULL, 0) & 0xffff;
		else
			return eval_operand(&operand[3]) & 0xffff;
	} else 
	if (strncmp(operand, "hi(", 3) == 0) {
		*strchr(operand, ')') = '\0';
		if ((operand[3] == '-') || isdigit(operand[3]))
			return (unsigned long)strtol(&operand[3], NULL, 0) >> 16;
		else
			return eval_operand(&operand[3]) >> 16;
	} else 
	if (!strstr(operand, "(")) 	/* not indirect but ...*/
		if (*operand == 'r') 	/* ... register direct */
			return eval_reg(operand);
		
		else 			/* ... rel. or abs. address */
			return eval_label(operand);
	else {				/* register indirect */
		int disp;		/* with possible displacement */
		char *regstr;
		unsigned int memaddr;
		
		disp = atoi(operand);	/* operand == "nn(rXX)" */
		debug("eval_operand: disp=%u", disp);
		regstr = strstr(operand, "(r") + 1;	/* regstr == "rXX)" */
		*strstr(regstr, ")") = '\0';		/* regstr == "rXX" */
		debug("eval_operand: regstr=%s", regstr);
		memaddr = eval_reg(regstr) + disp;
		
		return eval_mem32(memaddr);
	}
	
	return 0;
}

/* Implementation specific.
   Set destination operand (register direct, register indirect
   (with displacement) with value. */
   
void set_operand(char *dstoperand, unsigned long value)
{
	char operand[OPERANDNAME_LEN];
	
	debug("set_operand %s <= 0x%x (%d)\n", dstoperand, value, value);
	strcpy(operand, dstoperand);

	if (*operand == '#') {		/* immediate */
		printf("INTERNAL ERROR: Can't set immediate operand.\n");
		cont_run = 0;
	}
	else
	if (!strstr(operand, "(")) 	/* not indirect but ...*/
		if (*operand == 'r') 	/* ... register direct */
			set_reg32(operand, value);
		else {			/* ... rel. or abs. address */
			/* printf("INTERNAL ERROR: Can't set addr operand.\n"); 
			cont_run = 0; */
			set_mem32(eval_label(operand), value);
		}
	else {				/* register indirect */
		int disp;		/* with possible displacement */
		char *regstr;
		unsigned int memaddr;
		
		disp = atoi(operand);	/* operand == "nn(rXX)" */
		regstr = strstr(operand, "(r") + 1;	/* regstr == "rXX)" */
		*strstr(regstr, ")") = '\0';		/* regstr == "rXX" */
		memaddr = eval_reg(regstr) + disp;
		
		set_mem32(memaddr, value);
	}

	return;
}

void reset()
{
	cycles = 0;
	supercycles = 0;
	loadcycles = 0;
	storecycles = 0;
	memset(reg, 0, sizeof(reg));
	memset(iqueue, 0, sizeof(iqueue));
	memset(icomplet, 0, sizeof(icomplet));
	mtspr(SPR_SR, mfspr(SPR_SR) | SPR_SR_EXR);
	pctemp = eval_label("_main");
	pc = pctemp;
	debug("reset ...");
	set_reg32(STACK_REG, (MEMORY_START + MEMORY_LEN) - STACK_SIZE);
	except_handle(EXCEPT_RESET, 0);
}

void fetch()
{
	/* Cycles after reset. */
	cycles++;

	/* Simulate instruction cache */
	ic_simulate(pc);
	
	debug("fetch()\n");
	/* Fetch instruction. */
	strcpy(iqueue[0].insn, mem[pc].insn->insn);
	strcpy(iqueue[0].op1, mem[pc].insn->op1);
	strcpy(iqueue[0].op2, mem[pc].insn->op2);
	strcpy(iqueue[0].op3, mem[pc].insn->op3);
	strcpy(iqueue[0].op4, mem[pc].insn->op4);
	iqueue[0].insn_addr = pc;
	iqueue[0].dependdst = NULL;
	iqueue[0].dependsrc1 = NULL;
	iqueue[0].dependsrc2 = NULL;

	/* Increment program counter. */
	pc = pctemp;
	pctemp += 4;
	
	if ((pc < MEMORY_START) || (pc > MEMORY_START + MEMORY_LEN) || !mem[pc].insn) {
		except_handle(EXCEPT_BUSERR, pc);
		return;
	}
	
	/* Check for breakpoint. */
	if (mem[pc].brk)
		cont_run = 0;	/* Breakpoint set. */
	
	nop_period++;
	return;
}

void decode(struct iqueue_entry *cur)
{

	cur->dependdst = cur->op1;
	cur->dependsrc1 = cur->op2;	/* for calculating register */
	cur->dependsrc2 = cur->op3;	/* dependency */
	
	cur->func_unit = unknown;
	
	if (CURINSN("l.sw")) {
		cur->func_unit = store;
		set_operand(cur->op1, eval_operand(cur->op2));
	} else
	if (CURINSN("l.sb")) {
		cur->func_unit = store;
		set_operand(cur->op1, (eval_operand(cur->op2) << 24) + (eval_operand(cur->op1) & 0xffffff));
	} else
	if (CURINSN("l.sh")) {
		cur->func_unit = store;
		set_operand(cur->op1, (eval_operand(cur->op2) << 16) + (eval_operand(cur->op1) & 0xffff));
	} else
	if (CURINSN("l.lw")) {
		cur->func_unit = load;
		set_operand(cur->op1, eval_operand(cur->op2));
	} else
	if (CURINSN("l.lbs")) {
		signed char temp = (eval_operand(cur->op2) >> 24);
		cur->func_unit = load;
		set_operand(cur->op1, temp);
	} else
	if (CURINSN("l.lbz")) {
		unsigned char temp = (eval_operand(cur->op2) >> 24);
		cur->func_unit = load;
		set_operand(cur->op1, temp);
	} else
	if (CURINSN("l.lhs")) {
		signed short temp = (eval_operand(cur->op2) >> 16);
		cur->func_unit = load;
		set_operand(cur->op1, temp);
	} else
	if (CURINSN("l.lhz")) {
		unsigned short temp = (eval_operand(cur->op2) >> 16);
		cur->func_unit = load;
		set_operand(cur->op1, temp);
	} else
	if (CURINSN("l.movhi")) {
		cur->func_unit = movimm;
		set_operand(cur->op1, eval_operand(cur->op2) << 16);
	} else
	if (CURINSN("l.and") || CURINSN("l.andi")) {
		cur->func_unit = arith;
		set_operand(cur->op1, eval_operand(cur->op2) & (unsigned)eval_operand(cur->op3));
	} else
	if (CURINSN("l.or") || CURINSN("l.ori")) {
		cur->func_unit = arith;
		set_operand(cur->op1, eval_operand(cur->op2) | (unsigned)eval_operand(cur->op3));
	} else
	if (CURINSN("l.xor")) {
		cur->func_unit = arith;
		set_operand(cur->op1, eval_operand(cur->op2) ^ (signed)eval_operand(cur->op3));
	} else
	if (CURINSN("l.xori")) {
		cur->func_unit = arith;
		set_operand(cur->op1, eval_operand(cur->op2) ^ (signed)eval_operand(cur->op3));
	} else
	if (CURINSN("l.add")) {
		signed long temp3, temp2, temp1;
		signed char temp4;

		cur->func_unit = arith;
		temp3 = eval_operand(cur->op3);
		temp2 = eval_operand(cur->op2);
		temp1 = temp2 + temp3;
		set_operand(cur->op1, temp1);
		
		temp4 = temp1;
		if (temp4 == temp1)
			mstats.byteadd++;
	} else
	if (CURINSN("l.addi")) {
		signed long temp3, temp2, temp1;
		signed char temp4;
		
		cur->func_unit = arith;
		temp3 = eval_operand(cur->op3);
		temp2 = eval_operand(cur->op2);
		temp1 = temp2 + temp3;
		debug("l.addi: %x = %x + %x\n", temp1, temp2, temp3);
		set_operand(cur->op1, temp1);
		
		temp4 = temp1;
		if (temp4 == temp1)
			mstats.byteadd++;
	} else
	if (CURINSN("l.sub")) {
		signed long temp3, temp2, temp1;

		cur->func_unit = arith;
		temp3 = eval_operand(cur->op3);
		temp2 = eval_operand(cur->op2);
		temp1 = temp2 - temp3;
		set_operand(cur->op1, temp1);
	} else
	if (CURINSN("l.subi")) {
		signed long temp3, temp2, temp1;

		cur->func_unit = arith;
		temp3 = eval_operand(cur->op3);
		temp2 = eval_operand(cur->op2);
		temp1 = temp2 - temp3;
		set_operand(cur->op1, temp1);
	} else
	if (CURINSN("l.mul")) {
		signed long temp3, temp2, temp1;

		cur->func_unit = arith;
		temp3 = eval_operand(cur->op3);
		temp2 = eval_operand(cur->op2);
		temp1 = temp2 * temp3;
		set_operand(cur->op1, temp1);
	} else
	if (CURINSN("l.div")) {
		signed long temp3, temp2, temp1;

		cur->func_unit = arith;
		temp3 = eval_operand(cur->op3);
		temp2 = eval_operand(cur->op2);
		if (temp3)
			temp1 = temp2 / temp3;
		else
			except_handle(EXCEPT_ILLEGAL, 0);
		set_operand(cur->op1, temp1);
	} else
	if (CURINSN("l.divu")) {
		unsigned long temp3, temp2, temp1;

		cur->func_unit = arith;
		temp3 = eval_operand(cur->op3);
		temp2 = eval_operand(cur->op2);
		temp1 = temp2 / temp3;
		/* cycles += 16; */
		set_operand(cur->op1, temp1);
	} else
	if (CURINSN("l.sll") || CURINSN("l.slli")) {
		int sign = 1;
		cur->func_unit = shift;
		if ((signed)eval_operand(cur->op2) < 0)
			sign = -1;
		/* cycles += 2; */
		set_operand(cur->op1, eval_operand(cur->op2) << eval_operand(cur->op3));
	} else
	if (CURINSN("l.sra") || CURINSN("l.srai")) {
		unsigned long sign = 0;
		cur->func_unit = shift;
		
		if ((signed)eval_operand(cur->op2) < 0)
			sign = -1;
		/* cycles += 2; */
		set_operand(cur->op1, (signed)eval_operand(cur->op2) / (1 << eval_operand(cur->op3)));
	} else
	if (CURINSN("l.srl") || CURINSN("l.srli")) {
		cur->func_unit = shift;
		/* cycles += 2; */
		set_operand(cur->op1, eval_operand(cur->op2) >> eval_operand(cur->op3));
	} else
	if (CURINSN("l.j")) {
		char *endptr;
		strtol(cur->op1, &endptr, 0);
		if (*endptr == '\0') {
			debug("\nl.j relative: pc=%x pctemp=%x\n", pc, pctemp);
			pctemp = pc + (signed)eval_operand(cur->op1) * 4 - 4;
		} else
			pctemp = eval_operand(cur->op1);
		cur->func_unit = jump;
	} else
	if (CURINSN("l.jal")) {
		char *endptr;
		strtol(cur->op1, &endptr, 0);
		if (*endptr == '\0') {
			debug("\nl.jal relative: pc=%x pctemp=%x\n", pc, pctemp);
			pctemp = pc + (signed)eval_operand(cur->op1) * 4 - 4;
		} else
			pctemp = eval_operand(cur->op1);

		cur->func_unit = jump;
                set_reg32(LINK_REG, pc + 4);
                slp_func_entry();
	} else
	if (CURINSN("l.jalr")) {
		cur->func_unit = jump;
		pctemp = eval_operand(cur->op1);
                set_reg32(LINK_REG, pc + 4);
                slp_func_exit();
	} else
	if (CURINSN("l.jr")) {
		cur->func_unit = jump;
		pctemp = eval_operand(cur->op1);
	} else
	if (CURINSN("l.rfe")) {
		cur->func_unit = exception;
		pctemp = eval_operand(cur->op1) & ~0x3;
		mtspr(SPR_SR, (mfspr(SPR_SR) & ~0x3) | (eval_operand(cur->op1) & 0x3));
	} else
	if (CURINSN("l.nop")) {
		cur->func_unit = nop;
		if (nop_period > nop_maxperiod)
			nop_maxperiod = nop_period;
		nop_period = 0;
		nops++;
	} else
	if (CURINSN("l.bnf")) {
		cur->func_unit = branch;
		cur->dependsrc1 = ccr_flag;		
		if (eval_operand(cur->op1) >= pc)
			mstats.beqz.forward++;
		else
			mstats.beqz.backward++;
		mstats.sbp_bnf.all++;
		if (flag == 0) {
			char *endptr;

			strtol(cur->op1, &endptr, 0);
			if (*endptr == '\0') {
				debug("\nl.bnf relative: pc=%x pctemp=%x\n", pc, pctemp);
				pctemp = pc + (signed)eval_operand(cur->op1) * 4 - 4;
			} else
				pctemp = eval_operand(cur->op1);

			mstats.beqz.taken++;
			bpb_update(cur->insn_addr, 1);
			btic_update(pctemp);
		} else {
			if (eval_operand(cur->op1) >= pc)
				mstats.sbp_bnf.correct++;
			mstats.beqz.nottaken++;
			bpb_update(cur->insn_addr, 0);
			btic_update(pc);
		}
	} else
	if (CURINSN("l.bf")) {
		cur->func_unit = branch;
		cur->dependsrc1 = ccr_flag;
		if (eval_operand(cur->op1) >= pc)
			mstats.bnez.forward++;
		else
			mstats.bnez.backward++;
		mstats.sbp_bf.all++;
		if (flag) {
			char *endptr;

			strtol(cur->op1, &endptr, 0);
			if (*endptr == '\0') {
				debug("\nl.bf relative: pc=%x pctemp=%x\n", pc, pctemp);
				pctemp = pc + (signed)eval_operand(cur->op1) * 4 - 4;
			} else
				pctemp = eval_operand(cur->op1);

			if (eval_operand(cur->op1) < pc)
				mstats.sbp_bf.correct++;
			mstats.bnez.taken++;
			bpb_update(cur->insn_addr, 1);
			btic_update(pctemp);
		} else {
			mstats.bnez.nottaken++;
			bpb_update(cur->insn_addr, 0);
			btic_update(pc);
		}
	} else
	if (CURINSN("l.sfeq") || CURINSN("l.sfeqi")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = eval_operand(cur->op1) == eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfne") || CURINSN("l.sfnei")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = eval_operand(cur->op1) != eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfgts") || CURINSN("l.sfgtsi")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) >
		       (signed)eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfges") || CURINSN("l.sfgesi")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) >=
		       (signed)eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sflts") || CURINSN("l.sfltsi")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) <
		       (signed)eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfles") || CURINSN("l.sflesi")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (signed)eval_operand(cur->op1) <=
		       (signed)eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfgtu") || CURINSN("l.sfgtui")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) >
		       (unsigned)eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfgeu") || CURINSN("l.sfgeui")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) >=
		       (unsigned) eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfltu") || CURINSN("l.sfltui")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) <
		       (unsigned)eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.sfleu") || CURINSN("l.sfleui")) {
		cur->func_unit = compare;
		cur->dependsrc2 = cur->op1;
		cur->dependdst = ccr_flag;
		flag = (unsigned)eval_operand(cur->op1) <=
		       (unsigned)eval_operand(cur->op2);
		mtspr(SPR_CCR, mfspr(SPR_CCR) & ~SPR_CCR_FLAG | flag);
	} else
	if (CURINSN("l.mtsr")) {
		cur->func_unit = move;
		if (mfspr(SPR_SR) & SPR_SR_SUPV)
			mtspr(eval_operand(cur->op1), eval_operand(cur->op2));
	} else
	if (CURINSN("l.mfsr")) {
		cur->func_unit = move;
		if (mfspr(SPR_SR) & SPR_SR_SUPV)
			set_operand(cur->op1, mfspr(eval_operand(cur->op2)));
		else
			set_operand(cur->op1, 0);
	} else
	if (CURINSN("simrdtsc")) { /* obsolete */
		set_operand(cur->op1, cycles + loadcycles + storecycles);
	} else
	if (CURINSN("simprintf")) { /* obsolete */
		unsigned long stackaddr, fmtaddr, args;
		
		stackaddr = eval_reg(FRAME_REG);
		simprintf(stackaddr, eval_reg("r3"));
		debug("simprintf %x %x %x\n", stackaddr, fmtaddr, args);
	} else
	if (CURINSN("l.sys")) {
		if (eval_operand(cur->op1) > 200) {
			unsigned long stackaddr, fmtaddr, args;
			switch (eval_operand(cur->op1)) {
				case 201:
					set_operand("r3", cycles + loadcycles + storecycles);
					break;
				case 202:
					simprintf(stackaddr, eval_reg("r3"));
					debug("simprintf %x %x %x\n", stackaddr, fmtaddr, args);
					break;
				case 203:
					printf("syscall exit(%d)\n", eval_operand("r3"));
					cont_run = 0;
				default:
			}
		}
		else
			except_handle(EXCEPT_SYSCALL, 0);
	} else
		except_handle(EXCEPT_ILLEGAL, 0);

	/* Dynamic, dependency stats. */
	adddstats(icomplet[0].insn, iqueue[0].insn, 1, check_depend());

	/* Dynamic, functional units stats. */
	addfstats(icomplet[0].func_unit, iqueue[0].func_unit, 1, check_depend());

	/* Dynamic, single stats. */
	addsstats(iqueue[0].insn, 1, 0);

        if (cur->func_unit == store)
		storecycles += 0;
                        
        if (cur->func_unit == load)
                loadcycles += 0;

        if ((icomplet[0].func_unit == load) && check_depend())
                loadcycles++;

	/* Pseudo multiple issue benchmark */
	if ((multissue[cur->func_unit] < 1) || (check_depend())
	    || (issued_per_cycle < 1)
	   ) {
		int i;
		for (i = 0; i < 20; i++)
			multissue[i] = 2;
		issued_per_cycle = 2;
		supercycles++;
		if (check_depend())
			hazardwait++;
		multissue[unknown] = 2;
		multissue[shift] = 2;
		multissue[compare] = 1;
		multissue[branch] = 1;
		multissue[jump] = 1;
		multissue[extend] = 2;
		multissue[nop] = 2;
		multissue[move] = 2;
		multissue[movimm] = 2;
		multissue[arith] = 2;
		multissue[store] = 2;
		multissue[load] = 2;
	}
	multissue[cur->func_unit]--;
	issued_per_cycle--;
	
	return;
}

void execute()
{
	int i;
	
	/* Here comes real execution someday... */
	
	/* Instruction waits in completition buffer until retired. */
	strcpy(icomplet[0].insn, iqueue[0].insn);
	strcpy(icomplet[0].op1, iqueue[0].op1);
	strcpy(icomplet[0].op2, iqueue[0].op2);
	strcpy(icomplet[0].op3, iqueue[0].op3);
	strcpy(icomplet[0].op4, iqueue[0].op4);
	icomplet[0].func_unit = iqueue[0].func_unit;	
	icomplet[0].insn_addr = iqueue[0].insn_addr;

	if (iqueue[0].dependdst == iqueue[0].op1)
		icomplet[0].dependdst = icomplet[0].op1;
	else
	if (iqueue[0].dependdst == iqueue[0].op2)
		icomplet[0].dependdst = icomplet[0].op2;
	else
	if (iqueue[0].dependdst == iqueue[0].op3)
		icomplet[0].dependdst = icomplet[0].op3;
	else
		icomplet[0].dependdst = NULL;

	if (iqueue[0].dependsrc1 == iqueue[0].op1)
		icomplet[0].dependsrc1 = icomplet[0].op1;
	else
	if (iqueue[0].dependsrc1 == iqueue[0].op2)
		icomplet[0].dependsrc1 = icomplet[0].op2;
	else
	if (iqueue[0].dependsrc1 == iqueue[0].op3)
		icomplet[0].dependsrc1 = icomplet[0].op3;
	else
		icomplet[0].dependsrc1 = NULL;

	if (iqueue[0].dependsrc2 == iqueue[0].op1)
		icomplet[0].dependsrc2 = icomplet[0].op1;
	else
	if (iqueue[0].dependsrc2 == iqueue[0].op2)
		icomplet[0].dependsrc2 = icomplet[0].op2;
	else
	if (iqueue[0].dependsrc2 == iqueue[0].op3)
		icomplet[0].dependsrc2 = icomplet[0].op3;
	else
		icomplet[0].dependsrc2 = NULL;
	
	/* History of execution */
	for (i = HISTEXEC_LEN - 1; i; i--)
		histexec[i] = histexec[i - 1];
	histexec[0] = icomplet[0].insn_addr;	/* add last insn */

	/* Random interrupt generator. This is here just on a temporary basis. */
	if ((cycles % 1500 == 1400) && (rand() % 4 == 1))
		report_interrupt();
	
	return;
}

void dumpreg()
{
	int i;
	
	printf("\n\nIQ[0]:");
	dumpmemory(iqueue[0].insn_addr, iqueue[0].insn_addr + 4);
	printf(" (just executed)\tCYCLES: %u \nSuperscalar CYCLES: %u  HAZARDWAIT: %u\n", cycles, supercycles, hazardwait);
	printf("Additional LOAD CYCLES: %u  STORE CYCLES: %u\n", loadcycles, storecycles);
	printf("l.nop  count: %u  maxgap: %u\nPC:", nops, nop_maxperiod);
	dumpmemory(pc, pc + 4);
	printf(" (next insn)");
	for(i = 0; i < MAX_GPRS; i++) {
		if (i % 4 == 0)
			printf("\n");
		printf("GPR%.2u: %.8lx  ", i, reg[i]);
	}
	printf("flag: %u\n", flag);
}
